/**
 * \file MediaUtils.cpp
 * \brief Media related functions.
 * \author Marcus Overhagen, Francois Revol
 * Copyright 2003, OpenBeOS.
 */

#include <stdio.h>

#include <OS.h>
#include <Debug.h>
#include <math.h>
#include <MediaDefs.h>
#include <RealtimeAlloc.h>

#include "AudioRingBuffer.h"

/* fool BeIDE function parser */
#define IN_NAMESPACE \
namespace BPrivate { \
namespace Media { \
namespace Experimental {
IN_NAMESPACE
#undef IN_NAMESPACE

BAudioRingBuffer::BAudioRingBuffer(size_t size, bool rtAlloc=true, rtm_pool *fromRTPool=NULL)
	: BDataIO()
{
	initStatus = B_NO_MEMORY;
	fDataLen = size;
	fData = NULL;
	fPool = NULL;
	input_sem = 0;
	output_sem = 0;
	input_index = 0;
	output_index = 0;

	if (rtAlloc) { /* we want real-time alloc */
		fPool = fromRTPool;
		if (fPool == NULL) /* no pool specified, get the default */
			fPool = rtm_default_pool();
		if (fPool) { /* try to allocate */
			fData = rtm_alloc(fPool, fDataLen);
		}
	}
	if (fData == NULL) { /* no rt alloc, or not enough */
		fPool = NULL;
		fData = malloc(fDataLen);
	}
	if (fData == NULL)
		return;
	initStatus = B_NO_MORE_SEMS;
	input_sem = create_sem(fDataLen, "BAudioRingBuffer_input");
	if (input_sem < B_OK)
		return;
	output_sem = create_sem(0, "BAudioRingBuffer_output");
    if (output_sem < B_OK)
    	return;
    initStatus = B_OK;
}

BAudioRingBuffer::~BAudioRingBuffer()
{
	if (fPool)
		rtm_free(fData);
	else
		free(fData);
	if (input_sem)
		delete_sem(input_sem);
	if (output_sem)
		delete_sem(output_sem);
}

status_t BAudioRingBuffer::InitCheck() const
{
	return initStatus;
}

ssize_t BAudioRingBuffer::Read(void *buffer, size_t size)
{
	size_t len, amount;
	unsigned char *buf = (unsigned char *)buffer;
	unsigned char *dat = (unsigned char *)fData;
	status_t err;

	if (initStatus)
		return B_NO_INIT;
	while (size > 0) {
		//len = MIN(dataLen/2, size);
		len = size;
		while ((err = acquire_sem_etc(output_sem, len, B_CAN_INTERRUPT, 0LL)) < B_OK) {
			if (err = B_INTERRUPTED)
				continue;
			return err;
		}
		amount = MIN(len, (fDataLen - output_index));
		memcpy(buf, &(dat[output_index]), amount);
		output_index += amount;
		if (output_index >= fDataLen) {
			output_index %= fDataLen;
			memcpy(buf + amount, &(dat[output_index]), len - amount);
			output_index += len-amount;
			output_index %= fDataLen;
		}
		release_sem_etc(input_sem, len, 0);
		buf += len;
		size -= len;
	}
	return (buf - ((unsigned char *)buffer));
}

ssize_t BAudioRingBuffer::Write(const void *buffer, size_t size)
{
	size_t len, amount;
	unsigned char *buf = (unsigned char *)buffer;
	unsigned char *dat = (unsigned char *)fData;
	status_t err;

	if (initStatus)
		return B_NO_INIT;
	while (size > 0) {
		//len = MIN(dataLen/2, size);
		len = size;
		while ((err = acquire_sem_etc(input_sem, len, B_CAN_INTERRUPT, 0LL)) < B_OK) {
			if (err = B_INTERRUPTED)
				continue;
			return err;
		}
		amount = MIN(len, (fDataLen - input_index));
		memcpy(&(dat[input_index]), buf, amount);
		input_index += amount;
		if (input_index >= fDataLen) {
			input_index %= fDataLen;
			memcpy(&(dat[input_index]), buf + amount, len - amount);
			input_index += len-amount;
			input_index %= fDataLen;
		}
		release_sem_etc(output_sem, len, 0);
		buf += len;
		size -= len;
	}
	return (buf - ((unsigned char *)buffer));
}

size_t BAudioRingBuffer::Queued() const
{
	return (output_index - input_index) % fDataLen;
}

}; // namespace Experimental
}; // namespace Media
}; // namespace BPrivate
