/**
 * \file MediaUtils.cpp
 * \brief Media related functions.
 * \author Marcus Overhagen, Francois Revol
 * Copyright 2003, OpenBeOS.
 */

#include <stdio.h>

#include <OS.h>
#include <Debug.h>
#include <math.h>
#include <MediaDefs.h>

#include "MediaUtils.h"

namespace BPrivate {
namespace Media {
namespace utils {

#define PRINTF_SOME_MEDIA_TYPE(_type)	case _type: \
		fprintf(toFile, "%s\ttype =\t%ld (" #_type ");\n", prefix, format->type); \
		break

void fprint_media_format(FILE *toFile, media_format *format, const char *prefix)
{
	if (prefix == NULL)
		prefix = "";
	fprintf(toFile, "%s media_format {\n", prefix);
	switch (format->type) {
	PRINTF_SOME_MEDIA_TYPE(B_MEDIA_NO_TYPE);
	PRINTF_SOME_MEDIA_TYPE(B_MEDIA_UNKNOWN_TYPE);
	PRINTF_SOME_MEDIA_TYPE(B_MEDIA_VBL);
	PRINTF_SOME_MEDIA_TYPE(B_MEDIA_TIMECODE);
	PRINTF_SOME_MEDIA_TYPE(B_MEDIA_MIDI);
	PRINTF_SOME_MEDIA_TYPE(B_MEDIA_TEXT);
	PRINTF_SOME_MEDIA_TYPE(B_MEDIA_HTML);
	PRINTF_SOME_MEDIA_TYPE(B_MEDIA_MULTISTREAM);
	PRINTF_SOME_MEDIA_TYPE(B_MEDIA_PARAMETERS);
	PRINTF_SOME_MEDIA_TYPE(B_MEDIA_ENCODED_AUDIO);
	PRINTF_SOME_MEDIA_TYPE(B_MEDIA_ENCODED_VIDEO);
	PRINTF_SOME_MEDIA_TYPE(B_MEDIA_RAW_VIDEO);

	case B_MEDIA_RAW_AUDIO:
		fprintf(toFile, "%s\ttype =\t%ld (" "B_MEDIA_RAW_AUDIO" ");\n", prefix, format->type);
		fprintf(toFile, "%s\tu.raw_audio = {\n", prefix);
		fprintf(toFile, "%s\t\tframe_rate\t= %f;\n", prefix, format->u.raw_audio.frame_rate);
		fprintf(toFile, "%s\t\tchannel_count\t= %ld;\n", prefix, format->u.raw_audio.channel_count);
		const char *byteOrderStr = "unknown";
		switch (format->u.raw_audio.byte_order) {
			case B_MEDIA_BIG_ENDIAN: byteOrderStr = "B_MEDIA_BIG_ENDIAN"; break;
			case B_MEDIA_LITTLE_ENDIAN: byteOrderStr = "B_MEDIA_LITTLE_ENDIAN"; break;
		}
		fprintf(toFile, "%s\t\tbyte_order\t= %ld (%s);\n", prefix, format->u.raw_audio.byte_order, byteOrderStr);

		fprintf(toFile, "%s\t\tbuffer_size\t= %ld;\n", prefix, format->u.raw_audio.buffer_size);
		const char *audioFormat = "unknown";
		switch (format->u.raw_audio.format) {
			case media_raw_audio_format::B_AUDIO_FLOAT: audioFormat = "B_AUDIO_FLOAT"; break;
			case media_raw_audio_format::B_AUDIO_SHORT: audioFormat = "B_AUDIO_SHORT"; break;
			case media_raw_audio_format::B_AUDIO_INT: audioFormat = "B_AUDIO_INT"; break;
			case media_raw_audio_format::B_AUDIO_CHAR: audioFormat = "B_AUDIO_CHAR"; break;
			case media_raw_audio_format::B_AUDIO_UCHAR: audioFormat = "B_AUDIO_UCHAR"; break;
		}
		fprintf(toFile, "%s\t\tformat\t= %ld (%s);\n", prefix, format->u.raw_audio.format, audioFormat);
		fprintf(toFile, "%s } // media_format\n", prefix);
		break;
	}
}

#undef PRINTF_SOME_MEDIA_TYPE

// These two conversions seem to pop up all the time in media code.
// I guess it's the curse of microsecond resolution... ;-)
double
us_to_s(bigtime_t usecs)
{
	return (usecs / 1000000.0);
}

bigtime_t
s_to_us(double secs)
{
	return (bigtime_t) (secs * 1000000.0);
}

int
bytes_per_frame(
	const media_raw_audio_format & format)
{
	//	The media_raw_audio_format format constants encode the
	//	bytes-per-sample value in the low nybble. Having a fixed
	//	number of bytes-per-sample, and no inter-sample relationships,
	//	is what makes a format "raw".
	int bytesPerSample = format.format & 0xf;
	return bytesPerSample * format.channel_count;
}

int
frames_per_buffer(
	const media_raw_audio_format & format)
{
	// This will give us the number of full-sized frames that will fit
	// in a buffer. (Remember, integer division automatically rounds
	// down.)
	int frames = 0;
	if (bytes_per_frame(format) > 0) {
		frames = format.buffer_size / bytes_per_frame(format);
	}
	return frames;
}

bigtime_t
buffer_duration(
	const media_raw_audio_format & format)
{
	//	Figuring out duration is easy. We take extra precaution to
	//	not divide by zero or return irrelevant results.
	bigtime_t duration = 0;
	if (format.buffer_size > 0 && format.frame_rate > 0 && bytes_per_frame(format) > 0) {
		//	In these kinds of calculations, it's always useful to double-check
		//	the unit conversions. (Anyone remember high school physics?)
		//	bytes/(bytes/frame) / frames/sec
		//	= frames * sec/frames
		//	= secs                            which is what we want.
		duration = s_to_us((format.buffer_size / bytes_per_frame(format)) / format.frame_rate);
	}
	return duration;
}

bigtime_t
frames_duration(
	const media_raw_audio_format & format, int64 num_frames)
{
	//	Tells us how long in us it will take to produce num_frames,
	//	with the given format.
	bigtime_t duration = 0;
	if (format.frame_rate > 0) {
		duration = s_to_us(num_frames/format.frame_rate);
	}
	return duration;
}

int
buffers_for_duration(
	const media_raw_audio_format & format, bigtime_t duration)
{
	// Double-checking those unit conversions again:
	// secs * ( (frames/sec) / (frames/buffer) ) = secs * (buffers/sec) = buffers
	int buffers = 0;
	if (frames_per_buffer(format) > 0) {
		buffers = (int) ceil(us_to_s(duration)*(format.frame_rate/frames_per_buffer(format)));
	}
	return buffers;
}

int64
frames_for_duration(
	const media_raw_audio_format & format, bigtime_t duration)
{
	return (int64) ceil(format.frame_rate*us_to_s(duration));
}

}; // namespace utils
}; // namespace Media
}; // namespace BPrivate
