/**
 * \file SoundRecorderNode.h
 * \brief This is the BBufferConsumer, used internally by BSoundRecorder.
 * \author Marcus Overhagen, Francois Revol
 * Copyright 2003, OpenBeOS.
 */


#if !defined( _SOUND_RECORDER_NODE_h )
#define _SOUND_RECORDER_NODE_h

#include <BufferConsumer.h>

//	All incoming buffers and Media Kit requests arrive at a
//	media node in the form of messages (which are generally
//	dispatched for you by your superclasses' HandleMessage
//	implementations). Each message has a 'type' which is
//	analagous to a BMessage's 'what' field. We'll define our
//	own private message types for our SoundConsumer and
//	SoundProducer to use. The BeOS reserves a range,
//	0x60000000 to 0x7fffffff, for us to use.
enum {
	MSG_QUIT_NOW = 0x60000000L
};

//	These are special codes that we use in the Notify
//	function hook.
enum {
	B_WILL_START = 1,		//	performance_time
	B_WILL_STOP,			//	performance_time immediate
	B_WILL_SEEK,			//	performance_time media_time
	B_WILL_TIMEWARP,		//	real_time performance_time
	B_CONNECTED,			//	name (char*)
	B_DISCONNECTED,			//
	B_FORMAT_CHANGED,		//	media_raw_audio_format*
	B_NODE_DIES,			//	node will die!
	B_HOOKS_CHANGED,		//	
	B_OP_TIMED_OUT,			//	timeout that expired -- Consumer only
	B_PRODUCER_DATA_STATUS,	//	status performance_time -- Consumer only
	B_LATE_NOTICE			//	how_much performance_time -- Producer only
};
// FIXME: these duplicate from SoundRecorder.h

//	To use this Consumer:
//	1. Create Record and Notify hooks, or subclass _SoundRecorderNode
//	   if you'd rather use the inheritance hierarchy.
//	   * The Record function should do whatever you want to do
//	     when you receive a buffer.
//	   * The Notify function should handle whatever events
//	     you wish to handle (defined in SoundUtil.h).
//	2: Create an instance of _SoundRecorderNode, giving it the
//	   appropriate hook functions. Or, create an instance of an
//	   appropriate subclass if you've made one.
//	3: Register your new Consumer with the MediaRoster.
//	4: Connect your Consumer to some Producer.
//	5: Start or Stop the Consumer if your hook functions
//	   implement behavior for these kinds of events.
//	   Seek the Consumer to set the offset of the timestamps that
//	   your Record function will see.
//	6: When you're done, disconnect the Consumer, then delete it.

class BSoundRecorder;

namespace BPrivate {
namespace Media {
namespace Experimental {

class _SoundRecorderNode :
	public BBufferConsumer
{
public:
		virtual ~_SoundRecorderNode();
		_SoundRecorderNode(const char *name, /*media_multi_audio_format *format, */BSoundRecorder *recorder);

		//	The MediaNode interface
public:
virtual	port_id ControlPort() const;
virtual	BMediaAddOn* AddOn(
				int32 * internal_id) const;	/* Who instantiated you -- or NULL for app class */


protected:

virtual	void Start(
				bigtime_t performance_time);
virtual	void Stop(
				bigtime_t performance_time,
				bool immediate);
virtual	void Seek(
				bigtime_t media_time,
				bigtime_t performance_time);
virtual	void SetRunMode(
				run_mode mode);
virtual	void TimeWarp(
				bigtime_t at_real_time,
				bigtime_t to_performance_time);
virtual	void Preroll();
virtual	void SetTimeSource(
				BTimeSource * time_source);
virtual	status_t HandleMessage(
				int32 message,
				const void * data,
				size_t size);

		//	The BufferConsumer interface
virtual	status_t AcceptFormat(
				const media_destination & dest,
				media_format * format);
virtual	status_t GetNextInput( /* cookie starts at 0 */
				int32 * cookie,
				media_input * out_input);
virtual	void DisposeInputCookie(
				int32 cookie);
virtual	void BufferReceived(
				BBuffer * buffer);
virtual	void ProducerDataStatus(
				const media_destination & for_whom,
				int32 status,
				bigtime_t at_media_time);
virtual	status_t GetLatencyFor(
				const media_destination & for_whom,
				bigtime_t * out_latency,
				media_node_id * out_timesource);
virtual	status_t Connected(
				const media_source & producer,
				const media_destination & where,
				const media_format & with_format,
				media_input * out_input);
virtual	void Disconnected(
				const media_source & producer,
				const media_destination & where);
virtual	status_t FormatChanged(
				const media_source & producer,
				const media_destination & consumer, 
				int32 from_change_count,
				const media_format & format);

private:
friend class BSoundRecorder;
		BSoundRecorder *m_recorder;
//		media_multi_audio_format *m_format;
		media_input m_input;
		thread_id m_thread;
		port_id m_port;
		
		char *m_incomingMessage; // a message read by ServiceThread

		// The times we need to deal with
		// My notation for times: tr = real time,
		// tp = performance time, tm = media time.
		bigtime_t m_trTimeout;	// how long to wait on the input port
		bigtime_t m_tpSeekAt;	// when we Seek
		bigtime_t m_tmSeekTo;	// target time for Seek

		// The transformation from media to peformance time.
		// d = p - m, so m + d = p.
		// Media time is generally governed by the Seek
		// function. In our node, we simply use media time as
		// the time that we report to the record hook function.
		// If we were a producer node, we might use media time
		// to track where we were in playing a certain piece
		// of media. But we aren't.
		bigtime_t m_delta;
		
		// State variables
		bool m_seeking;		// a Seek is pending

		//	Functions to calculate timing values. OK to override.
		//	ProcessingLatency is the time it takes to process a buffer;
		//	TotalLatency is returned to producer; Timeout is passed
		//	to call to read_port_etc() in service thread loop.
virtual	bigtime_t Timeout();
virtual	bigtime_t ProcessingLatency();
virtual bigtime_t TotalLatency();
		//	The actual thread doing the work
static	status_t ThreadEntry(
				void * obj);
		void ServiceThread();
		void DoHookChange(
				void * msg);
};

} // Experimental
} // Media
} // BPrivate

#endif	/* _SOUND_RECORDER_NODE_h */
